#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# YMP documentation build configuration file

# This file is execfile()d with the current directory set to its
# containing dir.

import os
import sys
sys.path.insert(0, os.path.join(os.path.abspath(os.pardir), 'src'))

import ymp
import cloud_sptheme as csp

docdir = os.path.dirname(__file__)
ympdir = os.path.join(os.path.dirname(docdir), 'src')



# -- General configuration ------------------------------------------------

needs_sphinx = '1.6'


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinxcontrib.fulltoc',
    'sphinx_click.ext',
    'ymp.sphinxext',
]

autoclass_content = "both"
default_role = "any"

templates_path = ['_templates']
html_static_path = ['_static']
source_suffix = ['.rst']
master_doc = 'contents'  # master toc
index_doc = 'index'  # frontpage

project = 'YMP Extensible Omics Pipeline'
author = 'Elmar Pruesse'
copyright = '2017-2018, ' + author

version = ".".join(ymp.__version__.split(".")[:3])
release = ymp.__version__

language = None

exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store'
]
exclude_patterns.extend(templates_path)
exclude_patterns.extend(html_static_path)

pygments_style = 'sphinx'
todo_include_todos = True
keep_warnings = True
issue_tracker_url = 'gh:epruesse/ymp'

modindex_common_prefix = [ "ymp." ]

html_theme = "cloud"
html_theme_path = [csp.get_theme_dir()]
html_theme_options = {
    'roottarget': index_doc,
    'borderless_decor': True,
    'sidebarwidth': "3in",
    'hyphenation_language': 'en',
    #'rubricbgcolor': '#b3b3b3',
}
# html_title
# html_short_title
# html_logo
# html_favicon


# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
html_sidebars = {'**': ['searchbox.html', 'globaltoc.html']}

# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = 'YMPdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'letterpaper',
    'pointsize': '10pt',
    'preamble': '',
    'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'YMP.tex', 'YMP Documentation',
     'Elmar Pruesse', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'ymp', 'YMP Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'YMP', 'YMP Documentation',
     author, 'YMP', 'One line description of project.',
     'Miscellaneous'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'snakemake': ('http://snakemake.readthedocs.io/en/stable', None),
    'sphinx': ('http://www.sphinx-doc.org/en/stable', None),
    'bioconda': ('https://bioconda.github.io', None),
    'click': ('http://click.pocoo.org/5/', None),
}


def run_apidoc(_):
    """
    Calls sphinx-apidoc to generate template .rst files for each module
    """
    from sphinx.ext import apidoc
    apidoc.OPTIONS = ['members', 'show-inheritance']
    apidoc.main(['--output-dir', docdir,
                 '--doc-project', 'API',
                 '--force',
                 '--module-first',
                 os.path.join(ympdir, 'ymp')])


def setup(app):
    app.connect('builder-inited', run_apidoc)
    app.add_stylesheet('custom.css')



"Implements subcommands for ``ymp show``"

import logging

import click

import ymp
from ymp.cli.shared_options import command

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ConfigPropertyParam(click.ParamType):
    """Handles tab expansion for ``ymp show`` arguments"""
    _properties = None

    @property
    def properties(self):
        """Find properties offered by ConfigMgr"""
        if not self._properties:
            from ymp.config import ConfigMgr
            self._properties = {
                prop: getattr(getattr(ConfigMgr, prop), "__doc__")
                for prop in dir(ConfigMgr)
                if (prop[0] != "_"
                    and isinstance(getattr(ConfigMgr, prop), property))
            }
        return self._properties

    def complete(self, _ctx, incomplete):
        """Try to complete incomplete command

        This is executed on tab or tab-tab from the shell

        Args:
          ctx: click context object
          incomplete: last word in command line up until cursor

        Returns:
          list of words incomplete can be completed to
        """
        # This only list public properties, no member variables,
        # mainly because member variables can't have docstrings set.

        return [x for x in self.properties.keys()
                if x.startswith(incomplete)]

    def convert(self, value, param, ctx):
        """Convert value of param given context

        Args:
          value: string passed on command line
          param: click parameter object
          ctx:   click context object
        """
        return value

    def __repr__(self):
        props = "\n".join(
            "  {}: {}".format(p, self.properties[p].strip())
            for p in self.properties
            if self.properties[p]
        )
        return "\n".join(["Properties:", props])


def show_help(ctx, _param=None, value=True):
    """Display click command help"""
    if value:
        helpstr = [ctx.get_help(), '']
        arg_docs = [repr(param.type)
                    for param in ctx.command.params
                    if isinstance(param, click.Argument)]
        click.echo("\n".join(helpstr + arg_docs), color=ctx.color)
        ctx.exit()


@command(add_help_option=False)
@click.argument(
    "prop", nargs=1, metavar="PROPERTY", required=False,
    type=ConfigPropertyParam()
)
@click.option(
    "--help", "-h", callback=show_help, expose_value=False, is_flag=True
)
@click.option(
    "--source", "-s", is_flag=True,
    help="Show source"
)
@click.pass_context
def show(ctx, prop, source):
    """
    Show configuration properties
    """
    if not prop:
        show_help(ctx)

    log.debug(f"querying prop {prop}")
    obj = ymp.get_config()
    while prop:
        key, _, prop = prop.partition(".")
        key, _, slice_str = key.partition("[")
        obj = getattr(obj, key)

        if not slice_str:
            continue

        for subslice_str in slice_str[:-1].replace("][", ",").split(","):
            try:
                indices = [int(part) if part else None
                           for part in subslice_str.split(":")]
                if len(indices) > 3:
                    log.warning(f"Malformed slice string '{slice_str}'")
                    indices = indices[:3]
                if len(indices) == 1:
                    obj = obj[indices[0]]
                else:
                    obj = obj[slice(*indices)]
            except ValueError:
                obj = obj[subslice_str]

    try:
        output = obj.to_yaml(source)
    except AttributeError:
        output = str(obj)

    click.echo(output)



#!/usr/bin/env python3
import pandas
import os, sys
import logging

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def merge_alternative_implementation(out, files):
    prefix = os.path.commonprefix(files)
    suffix = os.path.commonprefix([x[::-1] for x in files])[::-1]
    s1 = len(prefix)
    s2 = -len(suffix)
    #log.debug("Prefix = {}, Suffix = {}".format(prefix, suffix))
    for filename in files:
        with open(filename) as stream:
            attrs = {}
            headers = []
            for line in stream:
                cols = line[1:].strip().split("\t")
                if len(cols) == 2:
                    attrs[cols[0]] = cols[1]
                else:
                    headers = cols
                    break
            #log.warning("Reading {}".format(filename))
            df = pandas.read_csv(stream, sep="\t",
                                 names=[x if x != "Cov" else filename[s1:s2] for x in headers],
                                 index_col = [x for x,y in enumerate(headers) if y != "Cov"],
                                 nrows=1000
                                 )
        if filename == files[0]:
            allf = df
        else:
            #log.warning("Merging...")
            allf = allf.join(df)
    allf.to_csv(out)


def merge(out, files, collect = None, ignore = None):
    if collect is None:
        collect = []
    if ignore is None:
        ignore = []
    prefix = os.path.commonprefix(files)
    suffix = os.path.commonprefix([x[::-1] for x in files])[::-1]
    s1 = len(prefix)
    s2 = -len(suffix)
    log.debug("Prefix = {}, Suffix = {}".format(prefix, suffix))

    names = [filename[s1:s2] for filename in files]
    fps = []
    try:
        fps = [open(filename, "rb", buffering=81920) for filename in files]
        outfp = open(out, "wb")
    except IOError:
        raise Exception("unable to open \"{}\"".format(filename))

    headers = None
    for fp in fps:
        cols = []
        while len(cols) <= 2:
            cols = fp.readline()[1:].strip().split(b"\t")

        if not headers:
            headers = cols
        if headers != cols:
            raise Exception("Headers in {} do not match: {} != {}".format(
                            filename, headers, cols))

    icollect = [ header in collect for header in headers ]
    iignore = [ header in ignore for header in headers ]
    imatch = [ not header and not collect for header, collect in zip(icollect, iignore) ]

    outfp.write(b",".join([ x for x, match in zip(headers, imatch) if match ]))
    outfp.write(b",")
    outfp.write(",".join(names).encode('ascii'))
    outfp.write(b"\n")

    for line in fps[0]:
        arr = line.rstrip().split(b"\t")
        res = [ x for x, collect in zip(arr, icollect) if collect ]
        match = [ x for x, match in zip(arr, imatch) if match ]
        assert len(arr) == len(headers)
        for fp in fps[1:]:
            arr2 = fp.readline().split(b"\t")
            res += [ x for x, collect in zip(arr2, icollect) if collect ]

        outfp.write(b",".join(match + res) + b"\n")

    outfp.close()
    for fp in fps:
        fp.close()


if __name__ == "__main__":
    #profile.run('merge(sys.argv[1], sys.argv[2:])')
    #profile.run('merge2(sys.argv[1], sys.argv[2:], collect=b"Cov")')
    merge(sys.argv[1], sorted(sys.argv[2:]), collect=b"Cov")



import logging

import pytest

import ymp
from ymp.projects import SQLiteProjectData, PandasProjectData

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

def pickled_sqlite_pd(cfg):
    import pickle
    obj = SQLiteProjectData(cfg)
    pkl = pickle.dumps(obj)
    return pickle.loads(pkl)


@pytest.fixture(params=[PandasProjectData, SQLiteProjectData, pickled_sqlite_pd])
def project_data(request, project_dir, saved_cwd):
    ymp.get_config().unload()
    cfg = ymp.get_config()
    return [request.param(prj.cfg.data) for prj in cfg.projects.values()]


@pytest.mark.parametrize("project", ["complex_data"], indirect=True)
def test_load_data(project_data):
    print(project_data)
    assert project_data[0].dump() == project_data[1].dump()


@pytest.mark.parametrize("project", ["complex_data"], indirect=True)
def test_project_data(project_data):
    obj = project_data[0]
    cols = ['fruit', 'skin_color', 'meat_color', 'size', 'texture', 'taste',
            'weight', 'case']
    assert obj.columns() == cols
    assert obj.identifying_columns() == ['fruit']
    assert obj.duplicate_rows('fruit') == []
    assert obj.duplicate_rows('skin_color') == ['green', 'red', 'red', 'green']
    assert obj.string_columns() == cols
    for row, test in zip(
            obj.rows(['fruit']),
                     ['apple', 'orange', 'grape', 'banana', 'cherry', 'melon']
            ):
        assert row[1] == test
    assert obj.get('fruit', 'apple', 'skin_color') == ['green']
    assert obj.column('fruit') == ['apple', 'orange', 'grape', 'banana', 'cherry', 'melon']
    assert obj.groupby_dedup(cols[1:]) == cols[1:3]



"""
Collection of shared utility classes and methods
"""
import logging
import os
from collections import Iterable

import ymp

log = logging.getLogger(__name__)


class AttrDict(dict):
    """
    AttrDict adds accessing stored keys as attributes to dict
    """
    def __getattr__(self, attr):
        try:
            return super().__getattr__(attr)
        except AttributeError:
            try:
                val = self[attr]
            except KeyError as e:
                raise AttributeError(e)
            if isinstance(val, dict):
                return AttrDict(val)
            else:
                return val

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            super().__setattr__(attr, value)
        else:
            raise NotImplementedError()


class MkdirDict(AttrDict):
    "Creates directories as they are requested"
    def __getattr__(self, attr):
        dirname = super().__getattr__(attr)
        if not os.path.exists(dirname):
            log.info("Creating directory %s", dirname)
            os.makedirs(dirname, exist_ok=True)
        return dirname


def parse_number(s=""):
    """Basic 1k 1m 1g 1t parsing.

    - assumes base 2
    - returns "byte" value
    - accepts "1kib", "1kb" or "1k"
    """
    multiplier = 1
    s = s.strip().upper().rstrip("BI")

    if s.endswith("K"):
        multiplier = 1024
    if s.endswith("M"):
        multiplier = 1024*1024
    if s.endswith("G"):
        multiplier = 1024*1024*1024
    if s.endswith("T"):
        multiplier = 1024*1024*1024*1024

    s = s.rstrip("KMGT")

    if "." in s:
        return float(s)*multiplier
    else:
        return int(s)*multiplier


def flatten(item):
    """Flatten lists without turning strings into letters"""
    if isinstance(item, str):
        yield item
    elif isinstance(item, Iterable):
        for item2 in item:
            yield from flatten(item2)
    else:
        yield item


def is_container(obj):
    """Check if object is container, considering strings not containers"""
    return not isinstance(obj, str) and isinstance(obj, Iterable)


def ensure_list(obj):
    """Wrap ``obj`` in a `list` as needed"""
    if obj is None:
        return []
    if isinstance(obj, str) or not isinstance(obj, Iterable):
        return [obj]
    return list(obj)


class Cache(object):
    def __init__(self, root):
        import sqlite3
        os.makedirs(os.path.join(root), exist_ok=True)
        self.conn = sqlite3.connect(os.path.join(root, "ymp.db"),
                                    check_same_thread=False)

        # Drop tables if the database has the wrong version number
        version = self.conn.execute("PRAGMA user_version").fetchone()[0]
        if version == ymp.__numeric_version__ and version > 0:
            try:
                curs = self.conn.execute("SELECT file, time from stamps")
                update = any(os.path.getmtime(row[0]) > row[1] for row in curs)
            except FileNotFoundError:
                update = True
            del curs
            if update:
                log.error("dropping cache")
                self.conn.executescript("""
                DROP TABLE caches;
                DROP TABLE stamps;
                """)
        else:
            update = True

        if update:
            self.conn.executescript("""
            BEGIN EXCLUSIVE;
            DROP TABLE IF EXISTS caches;
            CREATE TABLE caches (
                name TEXT,
                key TEXT,
                data,
                PRIMARY KEY (name, key)
            );
            DROP TABLE IF EXISTS stamps;
            CREATE TABLE stamps (
                file TEXT PRIMARY KEY,
                time INT
            );

            PRAGMA user_version={};
            COMMIT;
            """.format(ymp.__numeric_version__))

        self.caches = {}
        self.files = {}

    def close(self):
        self.conn.close()

    def get_cache(self, name, clean=False, *args, **kwargs):
        if name not in self.caches:
            self.caches[name] = CacheDict(self, name, *args, **kwargs)
        return self.caches[name]

    def store(self, cache, key, obj):
        import pickle

        files = ensure_list(getattr(obj, "defined_in", None))
        try:
            stamps = [(fn, os.path.getmtime(fn))
                      for fn in files
                      if fn not in self.files]
            self.conn.executemany(
                "REPLACE INTO stamps VALUES (?,?)",
                stamps)
            self.files.update(dict(stamps))
            self.conn.execute("""
              REPLACE INTO caches
              VALUES (?, ?, ?)
            """, [cache, key, pickle.dumps(obj)]
            )
        except FileNotFoundError:
            pass

    def commit(self):
        try:
            self.conn.commit()
        except sqlite3.OperationalError as e:
            log.warning("Cache write failed: %s", e.what())

    def load(self, cache, key):
        import pickle
        row = self.conn.execute("""
        SELECT data FROM caches WHERE name=? AND key=?
        """, [cache, key]).fetchone()
        if row:
            return pickle.loads(row[0])
        else:
            return None

    def load_all(self, cache):
        import pickle
        rows = self.conn.execute("""
        SELECT key, data FROM caches WHERE name=?
        """, [cache])
        return ((row[0], pickle.loads(row[1]))
                for row in rows)


class CacheDict(AttrDict):
    def __init__(self, cache, name, *args, loadfunc=None,
                 itemloadfunc=None, itemdata=None, **kwargs):
        self._cache = cache
        self._name = name
        self._loadfunc = loadfunc
        self._itemloadfunc = itemloadfunc
        self._itemdata = itemdata
        self._args = args
        self._kwargs = kwargs
        self._loading = False
        self._complete = False

    def _loaditem(self, key):
        cached = self._cache.load(self._name, key)
        if cached:
            super().__setitem__(key, cached)
        elif self._itemdata is not None:
            if key in self._itemdata:
                item = self._itemloadfunc(key, self._itemdata[key])
                self._cache.store(self._name, key, item)
                self._cache.commit()
                super().__setitem__(key, item)
        elif self._itemloadfunc:
            item = self._itemloadfunc(key)
            self._cache.store(self._name, key, item)
            self._cache.commit()
            super().__setitem__(key, item)
        else:
            self._loadall()

    def _loadall(self):
        if self._complete:
            return
        loaded = set()
        for key, obj in self._cache.load_all(self._name):
            loaded.add(key)
            super().__setitem__(key, obj)
        if self._itemloadfunc:
            for key in self._itemdata:
                if key not in loaded:
                    self._loaditem(key)
        elif self._loadfunc and not self._loading and not loaded:
            self._loadfunc(*self._args, **self._kwargs)
            self._loadfunc = None
            for key, item in super().items():
                self._cache.store(self._name, key, item)
            self._cache.commit()
        self._complete = True

    def __enter__(self):
        self._loading = True
        return self

    def __exit__(self, a, b, c):
        self._loading = False

    def __contains__(self, key):
        if self._itemdata:
            return key in self._itemdata
        self._loadall()
        return super().__contains__(key)

    def __len__(self):
        if self._itemdata:
            return len(self._itemdata)
        self._loadall()
        return super().__len__()

    def __getitem__(self, key):
        if not super().__contains__(key):
            self._loaditem(key)
        return super().__getitem__(key)

    def __setitem__(self, key, val):
        super().__setitem__(key, val)

    def __delitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        if self._itemdata:
            return self._itemdata.__iter__()
        self._loadall()
        return super().__iter__()

    def __str__(self):
        self._loadall()
        return super().__str__()

    def get(self, key, default=None):
        if not super().__contains__(key):
            self._loaditem(key)
        return super().get(key, default)

    def items(self):
        self._loadall()
        return super().items()

    def keys(self):
        if self._itemdata:
            return self._itemdata.keys()
        return super().keys()

    def values(self):
        self._loadall()
        return super().values()



#!/usr/bin/env python3

import fastentrypoints  # NOQA pylint: disable=unused-import
from setuptools import setup, find_packages


def read_file(fn):
    with open(fn) as f:
        content = f.read()
    return content


def get_content_type(fn):
    if fn.endswith(".md"):
        return "text/markdown"
    if fn.endswith(".rst"):
        return "text/x-rst"
    return "text/plain"


setup(
    name="ymp",
    use_scm_version={'write_to': 'src/ymp/_version.py'},
    description="Flexible multi-omic pipeline system",
    long_description=read_file("README.rst"),
    long_description_content_type=get_content_type("README.rst"),
    url="https://github.com/epruesse/ymp",
    author="Elmar Pruesse",
    author_email="elmar.pruesse@ucdenver.edu",
    license="GPL-3",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    platforms=["linux", "macos"],
    keywords=("bioinformatics pipeline rnaseq metagenomics "
              "conda bioconda snakemake"),
    project_urls={
        'Documentation': 'https://ymp.readthedocs.io',
        'Source': 'https://github.com/epruesse/ymp',
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    zip_safe=False,
    setup_requires=[
        'setuptools_scm>=3.2',
        'pytest-runner',
    ],
    tests_require=[
        'pytest-xdist',
        'pytest-logging',
        'pytest-timeout',
        'pygraphviz',
        'pytest',
        'yappi',
        'xlrd',
    ],
    install_requires=[
        'snakemake>=5.4.4',
        'Click',
        'Click-completion',
        'ruamel.yaml>0.15',
        'drmaa',
        'pandas>=0.20',
        'networkx>=2',
        'coloredlogs',
        'xdg',
        'aiohttp',
        'tqdm>=4.21.0',
    ],
    extras_require={
        'docs': [
            'sphinx',
            'cloud_sptheme',
            'sphinxcontrib-fulltoc',
            'sphinx-click',
            'sphinx_autodoc_typehints',
            'ftputil',
        ]
    },
    python_requires='>=3.6',
    include_package_data=True,
    entry_points='''
        [console_scripts]
        ymp=ymp.cli:main
    ''',
)



"""
Parsers for blast output formats 6 (CSV) and 7 (CSV with comments between queries).
"""

from collections import namedtuple
from typing import List


def reader(fileobj, t: int=7) -> 'BlastParser':
    """
    Creates a reader for files in BLAST format

    >>> with open(blast_file) as infile:
    >>>    reader = blast.reader(infile)
    >>>    for hit in reader:
    >>>       print(hit)

    Args:
      fileobj: iterable yielding lines in blast format
      t: number of blast format type
    """
    if t == 7:
        return Fmt7Parser(fileobj)
    elif t == 6:
        return Fmt6Parser(fileobj)
    else:
        ValueError("other formats not implemented")


class BlastParser(object):
    "Base class for BLAST parsers"

    # Map between field short and long names
    FIELD_MAP = {
        "query acc.": "qacc",
        "subject acc.": "sacc",
        "% identity": "pident",
        "alignment length": "length",
        "mismatches": "mismatch",
        "gap opens": "gapopen",
        "q. start": "qstart",
        "q. end": "qend",
        "s. start": "sstart",
        "s. end": "send",
        "evalue": "evalue",
        "bit score": "bitscore",
        "subject strand": "sstrand",
        "sbjct frame": "sframe",
        "score": "score"
    }

    # Map defining types of fields
    FIELD_TYPE = {
        'pident': float,
        'length': int,
        'mismatch': int,
        'gapopen': int,
        'qstart': int,
        'qend': int,
        'sstart': int,
        'send': int,
        'evalue': float,
        'bitscore': float,
        'score': float,
        'sframe': int
    }


class Fmt7Parser(BlastParser):
    """
    Parses BLAST results in format '7' (CSV with comments)
    """
    FIELDS = "# Fields: "
    QUERY = "# Query: "
    DATABASE = "# Database: "
    HITSFOUND = " hits found"

    def __init__(self, fileobj):
        self.fileobj = fileobj
        self.fields = None
        if "BLAST" not in fileobj.readline():
            raise ValueError("not a BLAST7 formatted file")

    def get_fields(self) -> List[str]:
        """Returns list of available field names

        Format 7 specifies which columns it contains in comment lines, allowing
        this parser to be agnostic of the selection of columns made when running
        BLAST.

        Returns:
          List of field names (e.g. ``['sacc', 'qacc', 'evalue']``)
        """
        return self.fields

    def __iter__(self):
        for line in self.fileobj:
            if line.startswith(self.FIELDS):
                self.fields = [
                    self.FIELD_MAP[field]
                    if field in self.FIELD_MAP else field
                    for field in line[len(self.FIELDS):].strip().split(", ")
                ]
                self.Hit = namedtuple("BlastHit", self.fields)
            elif line.startswith(self.QUERY):
                self.query = line[len(self.QUERY):].strip()
            elif line.startswith(self.DATABASE):
                self.query = line[len(self.DATABASE):].strip()
            elif line.strip().endswith(self.HITSFOUND):
                self.hits = int(line.split()[1])
                self.hit = 0
            elif line[0] == "#":
                continue
            else:
                self.hit += 1
                yield self.Hit(*[
                    self.FIELD_TYPE[key](value)
                    if key in self.FIELD_TYPE else value
                    for key, value in zip(self.fields,
                                          line.strip().split('\t'))
                ])

    def isfirsthit(self) -> bool:
        """Returns `True` if the current hit is the first hit for the current
        query"""
        return self.hit == 1


class Fmt6Parser(BlastParser):
    """Parser for BLAST format 6 (CSV)
    """
    #: Default field types
    fields = ("qseqid sseqid pident length mismatch gapopen "
              "qstart qend sstart send evalue bitscore").split()
    field_types = [BlastParser.FIELD_TYPE.get(n, None) for n in fields ]
    Hit = namedtuple("BlastHit", fields)
    def __init__(self, fileobj):
        self.fileobj = fileobj

    def get_fields(self):
        return self.fields

    def __iter__(self):
        for line in self.fileobj:
            yield self.Hit(*[t(v) if t else v
                             for v, t in zip(line.split("\t"),
                                             self.field_types)])




"""
YMP processes data in stages, each of which is contained in its own directory.

.. code-block:: snakemake

  with Stage("trim_bbmap") as S:
    S.doc("Trim reads with BBMap")
    rule bbmap_trim:
      output: "{:this:}/{sample}{:pairnames:}.fq.gz"
      input:  "{:prev:}/{sample}{:pairnames:}.fq.gz"
      ...

"""

import logging
import re
from typing import TYPE_CHECKING

import ymp
from ymp.exceptions import YmpException, YmpRuleError, YmpStageError
from ymp.snakemake import ColonExpander, ExpandLateException, WorkflowObject
from ymp.string import PartialFormatter


if TYPE_CHECKING:
    from typing import List
    from snakemake.rules import Rule

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def norm_wildcards(pattern):
    for n in ("{target}", "{source}", "{:target:}"):
        pattern = pattern.replace(n, "{sample}")
    for n in ("{:targets:}", "{:sources:}"):
        pattern = pattern.replace(n, "{:samples:}")
    return pattern


class StageStack(object):
    """The "head" of a processing chain - a stack of stages
    """

    @classmethod
    def get(cls, path, stage=None):
        """
        Cached access to StageStack

        Args:
          path: Stage path
          stage: Stage object at head of stack
        """
        cfg = ymp.get_config()
        cache = cfg.cache.get_cache(cls.__name__, itemloadfunc=StageStack)
        return cache[path]

    def __str__(self):
        return self.path

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.stage})"

    def __init__(self, path, stage=None):
        self.name = path
        # immediate predecessors (sources) for the Stage of this Stack
        self.prevs = []

        cfg = ymp.get_config()
        registry = Stage.get_registry()
        stage_names = path.split(".")

        # project for this stage stack
        self.project = cfg.projects.get(stage_names[0])
        if not self.project:
            raise YmpStageError(f"No project for stage stack {path} found")

        top_stage = stage_names.pop()
        if stage:
            if not stage.match(top_stage):
                raise YmpStageError(
                    f"Internal error: {top_stage} not matched by {stage}")
        if not stage:
            stage = self._find_stage(top_stage)
        self.stage = stage

        self.path = getattr(stage, "dir", path)

        self.group = getattr(stage, "group", None)
        if stage_names and stage_names[-1].startswith("group_"):
            self.group = [stage_names.pop().split("_")[1]]

        # gather prev stage stacks for this stack, going backwards
        # through stack until all inputs have been satisfied
        inputs = set(self.stage.inputs)
        while stage_names and inputs:
            path = ".".join(stage_names)
            stage = self._find_stage(stage_names.pop())
            if not stage:
                continue
            stack = self.get(path, stage)
            provides = inputs.intersection(stack.outputs)
            if provides:
                inputs -= provides
                self.prevs.append(stack)

        if inputs:
            # Can't find the right types, try to present useful error message:
            words = []
            for suffix in inputs:
                words.extend((suffix, "--"))
                words.extend([name for name, stage in registry.items()
                              if suffix in stage.outputs])
                words.extend('\n')
            text = ' '.join(words)

            raise YmpStageError(
                f"""
                File type(s) '{" ".join(inputs)}' required by '{self.stage}'
                not found in '{self.name}'. Stages providing missing
                file types:
                {text}
                """
            )

        if self.group is None:
            groups = list(dict.fromkeys(group
                                        for p in reversed(self.prevs)
                                        for group in p.group))
            self.group = self.project.minimize_variables(groups)

        log.info("Stage stack %s using column %s", self, self.group)

    def _find_stage(self, name):
        cfg = ymp.get_config()
        registry = Stage.get_registry()

        if name.startswith("group_"):
            return None  ## fails FIXME
        if name.startswith("ref_"):
            refname = name[4:]
            if refname in cfg.ref:
                return cfg.ref[refname]
            else:
                raise YmpStageError(f"Unknown reference '{cfg.ref[refname]}'")
        for stage in registry.values():
            if stage.match(name):
                return stage
        raise YmpStageError(f"Unknown stage '{name}'")

    def complete(self, incomplete):
        registry = Stage.get_registry()
        cfg = ymp.get_config()
        result = []
        groups = ("group_" + name for name in self.project.variables)
        result += (opt for opt in groups if opt.startswith(incomplete))
        refs = ("ref_" + name for name in cfg.ref)
        result += (opt for opt in refs if opt.startswith(incomplete))
        for stage in registry.values():
            for name in (stage.name, stage.altname):
                if name and name.startswith(incomplete):
                    try:
                        self.get(".".join((self.path, name)))
                        result.append(name)
                    except YmpStageError:
                        pass

        return result

    @property
    def outputs(self):
        return self.stage.outputs

    @property
    def defined_in(self):
        return None

    def prev(self, args=None, kwargs=None):
        """
        Directory of previous stage
        """
        _, _, suffix = kwargs.get('item').partition("{:prev:}")
        if not kwargs or "wc" not in kwargs:
            raise ExpandLateException()
        suffix = norm_wildcards(suffix)

        for stack in self.prevs:
            if suffix in stack.outputs:
                return stack

    @property
    def targets(self):
        """
        Returns the current targets
        """
        return self.project.get_ids(self.group)

    def target(self, args, kwargs):
        """Finds the target in the prev stage matching current target"""
        prev = self.get(self.prev(args, kwargs).name)
        cur_target = kwargs['wc'].target
        return self.project.get_ids(prev.group, self.group, cur_target)


class Stage(WorkflowObject):
    """
    Creates a new stage

    While entered using ``with``, several stage specific variables
    are expanded within rules:

    * ``{:this:}`` -- The current stage directory
    * ``{:that:}`` -- The alternate output stage directory
    * ``{:prev:}`` -- The previous stage's directory

    """

    active = None
    """Currently active stage ("entered")"""

    def __init__(self, name: str, altname: str=None,
                 env: str=None, doc: str=None) -> None:
        """
        Args:
            name: Name of this stage
            altname: Alternate name of this stage (used for stages with
                multiple output variants, e.g. filter_x and remove_x.
            doc: See `Stage.doc`
            env: See `Stage.env`
        """
        # Stage name
        self.name: str = name
        # Alternate stage name
        self.altname: str = altname
        super().__init__()
        self.register()
        # Rules in this stage
        self.rules: List[Rule] = []
        # Stage Parameters
        self.params: List[Param] = []
        # Input / Output types
        self.inputs = set()
        self.outputs = set()
        # Regex matching self
        self._regex = None

        self.doc(doc or "")
        self.env(env)

    def doc(self, doc: str) -> None:
        """Add documentation to Stage

        Args:
          doc: Docstring passed to Sphinx
        """
        self.docstring = doc

    def env(self, name: str) -> None:
        """Add package specifications to Stage environment

        Note:
          This sets the environment for all rules within the stage,
          which leads to errors with Snakemake rule types
          not supporting conda environments

        Args:
          name: Environment name or filename

        >>> Env("blast", packages="blast =2.7*")
        >>> with Stage("test") as S:
        >>>    S.env("blast")
        >>>    rule testing:
        >>>       ...

        >>> with Stage("test", env="blast") as S:
        >>>    rule testing:
        >>>       ...

        >>> with Stage("test") as S:
        >>>    rule testing:
        >>>       conda: "blast"
        >>>       ...
        """
        self.conda_env = name

    def __enter__(self):
        if Stage.active is not None:
            raise YmpRuleError(
                self,
                f"Failed to enter stage '{self.name}', "
                f"already in stage {self.active.name}'."
            )

        Stage.active = self
        return self

    def __exit__(self, *args):
        Stage.active = None

    def __str__(self):
        if self.altname:
            return "|".join((self.name, self.altname))
        return self.name

    def __repr__(self):
        return (f"{self.__class__.__name__} {self!s} "
                f"({self.filename}:{self.lineno})")

    def _add_rule(self, rule):
        rule.ymp_stage = self
        # FIXME: disabled because it breaks pickling of Stage
        # self.rules.append(rule)

    def add_param(self, key, typ, name, value=None, default=None):
        """Add parameter to stage

        Example:
            >>> with Stage("test") as S
            >>>   S.add_param("N", "int", "nval", default=50)
            >>>   rule:
            >>>      shell: "echo {param.nval}"

            This would add a stage "test", optionally callable as "testN123",
            printing "50" or in the case of "testN123" printing "123".

        Args:
          char: The character to use in the Stage name
          typ:  The type of the parameter (int, flag)
          param: Name of parameter in params
          value: value ``{param.xyz}`` should be set to if param given
          default: default value for ``{{param.xyz}}`` if no param given
        """
        if typ == 'flag':
            self.params.append(ParamFlag(self, key, name, value, default))
        elif typ == 'int':
            self.params.append(ParamInt(self, key, name, value, default))
        elif typ == 'choice':
            self.params.append(ParamChoice(self, key, name, value, default))
        else:
            raise YmpRuleError(self, f"Unknown Stage Parameter type '{typ}'")

    def wc2path(self, wc):
        wildcards = self.wildcards
        for p in self.params:
            wildcards = wildcards.replace(p.constraint, "")
        return wildcards.format(**wc)

    def stack(self, wc):
        return StageStack.get(self.wc2path(wc), self)

    def match(self, name):
        if not self._regex:
            if self.altname:
                sname = "(" + "|".join((self.name, self.altname)) + ")"
            else:
                sname = self.name
            pat = sname + "".join(p.regex for p in self.params)
            self._regex = re.compile(pat)
        return self._regex.fullmatch(name) is not None

    def prev(self, args, kwargs):
        """Gathers {:prev:} calls from rules"""
        prefix, _, suffix = kwargs.get('item').partition("{:prev:}")
        if not prefix:
            self.inputs.add(norm_wildcards(suffix))
        return None

    def this(self, args=None, kwargs=None):
        """
        Directory of current stage
        """
        if not Stage.active:
            raise YmpException(
                "Use of {:this:} requires active Stage"
            )
        prefix, _, suffix = kwargs.get('item').partition("{:this:}")
        if not prefix:
            self.outputs.add(norm_wildcards(suffix))

        item = kwargs.get('item')
        if item is None:
            raise YmpException("Internal Error")
        prefix, _, pattern = item.partition("{:this:}")

        return self.wildcards

    @property
    def that(self):
        """
        Alternate directory of current stage

        Used for splitting stages
        """
        if not Stage.active:
            raise YmpException(
                "Use of {:that:} requires active Stage"
            )
        if not Stage.active.altname:
            raise YmpException(
                "Use of {:that:} requires with altname"
            )
        return "".join(["{_YMP_DIR}", self.altname] +
                       [p.pattern for p in self.params])

    @property
    def wildcards(self):
        return "".join(["{_YMP_DIR}", self.name] +
                       [p.pattern for p in self.params])


class StageExpander(ColonExpander):
    """
    Registers rules with stages when they are created
    """
    def expand_ruleinfo(self, rule, item, expand_args, rec):
        if not Stage.active:
            return item
        stage = Stage.active

        stage._add_rule(rule)
        if not item.conda_env and stage.conda_env:
            item.conda_env = stage.conda_env

        if stage.params:
            if not item.params:
                item.params = ((), {})
            for param in stage.params:
                item.params[1][param.name] = param.param_func()

        return super().expand_ruleinfo(rule, item, expand_args, rec)

    def expand_str(self, rule, item, expand_args, rec, cb):
        if cb:
            stage = Stage.active
            Stage.active = rule.ymp_stage
        expand_args['item'] = item
        val = super().expand_str(rule, item, expand_args, rec, cb)
        if cb:
            Stage.active = stage
        return val

    def expands_field(self, field):
        return field not in 'func'

    class Formatter(ColonExpander.Formatter, PartialFormatter):
        def get_value(self, key, args, kwargs):
            try:
                return self.get_value_(key, args, kwargs)
            except Exception as e:
                if not isinstance(e, ExpandLateException):
                    log.warning(f"key={key} args={args} kwargs={kwargs}",
                                exc_info=True)
                raise

        def get_value_(self, key, args, kwargs):
            stage = Stage.active
            if hasattr(stage, key):
                val = getattr(stage, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val
            if "wc" not in kwargs:
                raise ExpandLateException()
            stack = stage.stack(kwargs['wc'])
            if hasattr(stack, key):
                val = getattr(stack, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val
            if hasattr(stack.project, key):
                val = getattr(stack.project, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val
            return super().get_value(key, args, kwargs)


class Param(object):
    """Stage Parameter (base class)"""
    def __init__(self, stage, key, name, value=None, default=None):
        self.stage = stage
        self.key = key
        self.name = name
        self.value = value
        self.default = default

        self.wildcard = f"_yp_{self.name}"

    @property
    def constraint(self):
        if self.regex:
            return "," + self.regex
        return ""

    @property
    def pattern(self):
        """String to add to filenames passed to Snakemake

        I.e. a pattern of the form ``{wildcard,constraint}``
        """
        return f"{{{self.wildcard}{self.constraint}}}"


class ParamFlag(Param):
    """Stage Flag Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.value:
            self.value = self.key

        self.regex = f"({self.key}?)"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            if getattr(wildcards, self.wildcard, None):
                return self.value
            else:
                return ""
        return name2param


class ParamInt(Param):
    """Stage Int Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.default:
            raise YmpRuleError(
                self.stage,
                f"Stage Int Parameter must have 'default' set")

        self.regex = f"({self.key}\d+|)"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            val = getattr(wildcards, self.wildcard, None)
            if val:
                return val[len(self.key):]
            else:
                return self.default
        return name2param


class ParamChoice(Param):
    """Stage Choice Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default is not None:
            self.value += [""]
        self.regex = f"({self.key}({'|'.join(self.value)}))"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            val = getattr(wildcards, self.wildcard, None)
            if val:
                return val[len(self.key):]
            else:
                return self.default
        return name2param



"Implements subcommands for ``ymp init``"

import logging
import os
import shutil
import subprocess as sp

import click

import ymp
from ymp.cli.shared_options import command, group

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


@group()
def init():
    """Initialize YMP workspace"""


def have_command(cmd):
    try:
        sp.run("command -v " + cmd, shell=True, stdout=sp.DEVNULL, check=True)
    except sp.CalledProcessError:
        log.debug("Command '%s' not found", cmd)
        return False
    log.debug("Command '%s' found", cmd)
    return True


@init.command()
@click.option("--yes", "-y", is_flag=True, help="Confirm every prompt")
def cluster(yes):
    """Set up cluster"""
    cfg = ymp.get_config()._config

    if cfg.cluster.profile is not None and not yes:
        click.confirm("Cluster profile '{}' already configured. "
                      "Do you want to overwrite this setting?"
                      "".format(cfg.cluster.profile),
                      abort=True)

    log.warning("Trying to detect cluster type")
    log.debug("Checking for SLURM")
    if have_command("sbatch"):
        log.warning("Found SLURM. Updating config.")
        cfg["cluster"]["profile"] = "slurm"
    elif have_command("qsub"):
        log.warning("Detected SGE or Torque")
    else:
        log.warning("No cluster submit commands found")
        cfg["cluster"]["profile"] = None
    log.warning("Saving config")
    cfg.save()


@init.command()
@click.argument("name", required=False)
@click.option("--yes", "-y", is_flag=True, help="Confirm every prompt")
def project(name, yes):
    cfg = ymp.get_config()._config

    if not name:
        name = click.prompt("Please enter a name for the new project",
                            type=str)

    if name in cfg.projects and not yes:
        click.confirm("Project '{}' already configured. "
                      "Do you want to overwrite this project?"
                      "".format(name),
                      abort=True)

    cfg.projects[name] = {'data': None}
    log.warning("Saving config")
    cfg.save()


@init.command()
def demo():
    """
    Copies YMP tutorial data into the current working directory
    """
    click.echo("Copying tutorial data to current working directory...")
    cwd_path = os.getcwd()
    cwd_files = os.listdir(cwd_path)
    demo_path = os.path.join(ymp._rsc_dir, "demo")
    demo_files = os.listdir(demo_path)
    conflicts = [f for f in demo_files if f in cwd_files]
    if len(cwd_files) > 10:
        click.confirm(
            "WARNING: "
            "The current working directory already contains a lot of files.\n"
            "  Using an empty directory to start with is highly suggested.\n"
            "  Do you want to continue?", abort=True)
    if conflicts:
        click.confirm(
            "WARNING: "
            "This operation would overwrite the following files: {}\n"
            "  Do you want to continue?".format(conflicts), abort=True)
        for f in conflicts:
            if os.path.isdir(f):
                shutil.rmtree(f)
            else:
                os.unlink(f)
    for f in demo_files:
        src = os.path.join(demo_path, f)
        dst = os.path.join(cwd_path, f)
        if os.path.isdir(src):
            shutil.copytree(src, dst)
        else:
            shutil.copy(src, dst)
    click.echo("done.")
    click.echo("")
    click.echo("Try running 'ymp make toy.assemble_megahit.map_bbmap',")
    click.echo("or see https://ymp.readthedocs.io for more examples.")



#!/usr/bin/env python3

import click


AA = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
NU = 'TCAG'
B2N = {a: b for a, b in zip(NU, range(len(NU)))}


def nuc2num(seq):
    return sum([
        len(NU) ** pos * B2N[nuc]
        for pos, nuc in enumerate(reversed(seq))
    ])


def nuc2aa(seq):
    return ''.join([
        AA[nuc2num(codon)]
        for codon in zip(*[iter(seq)]*3)
    ])


@click.command()
@click.argument('input', type=click.File('r'))
@click.argument('output', type=click.File('w'))
def click_fasta_dna2aa(input, output):
    if input.name.endswith(".gz"):
        import gzip
        input = gzip.open(input.name, "rt")
    fasta_dna2aa(input, output)


def fasta_dna2aa(inf, outf):
    def write_aa(header, seq):
        # outf.write(header.encode('ascii'))
        outf.write(header)
        aa = nuc2aa(seq)
        if "start_type=GTG" in header:
            aa = 'M'+aa[1:]
        outf.write(('\n'.join([
            aa[s:s+60]
            for s in range(0, len(aa)+59, 60)
            ]).strip()+'\n')
            # .encode('ascii')
        )

    header = None
    seq = ""

    for line in inf:
        # line = line.decode('ascii')
        if line[0] == '>':
            if header:
                write_aa(header, seq)
            header = line
            seq = ""
        else:
            seq += line.strip()
    if header:
        write_aa(header, seq)


if __name__ == "__main__":
    # pylint does not get click decorators, disable warning:
    # pylint: disable=no-value-for-parameter
    click_fasta_dna2aa()



import logging
import textwrap
from fnmatch import fnmatch

import click

from ymp.cli.shared_options import group

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def wrap(header, data):
    wrapper = textwrap.TextWrapper(
        initial_indent=header + " ",
        subsequent_indent=" " * len(header) + " "
    )
    return "\n"+"\n".join(wrapper.wrap(" ".join(data)))


@group()
def stage():
    """
    Manipulate YMP stages
    """


@stage.command(name="list")
@click.option(
    "--long", "-l", "long_opt", is_flag=True,
    help="Show full stage descriptions"
)
@click.option(
    "--short", "-s", "short_opt", is_flag=True,
    help="Show only stage names"
)
@click.option(
    "--code", "-c", "code_opt", is_flag=True,
    help="Show definition file name and line number"
)
@click.option(
    "--types", "-t", "type_opt", is_flag=True,
    help="Show input/output types"
)
@click.argument(
    "stage_opt", metavar="STAGE", nargs=-1
)
def ls(long_opt, short_opt, stage_opt, code_opt, type_opt):
    """
    List available stages
    """
    if long_opt and short_opt:
        raise click.UsageError(
            "Options --long and --short are mutually exclusive")

    from ymp.stage import Stage
    all_stages = Stage.get_registry()
    if stage_opt:
        stages = [all_stages[m] for m in all_stages
                  if any(fnmatch(m, pat) for pat in stage_opt)]
    else:
        stages = list(all_stages.values())
    stages = sorted(list(set(stages)), key=lambda s: s.name)

    if not stages:  # nothing to show
        return

    name_width = max(len(x.name) for x in stages)
    for stage in stages:
        if hasattr(stage, 'docstring'):
            doc = stage.docstring.strip().split("\n", 1)
            short_doc = doc[0].strip()
            if len(doc) > 1:
                long_doc = textwrap.dedent(doc[1])
            else:
                long_doc = ""
        else:
            short_doc = "[no docs]"
            long_doc = ""

        if long_doc and long_opt:
            description = "\n".join("  " + l for l in long_doc.split("\n"))
        else:
            description = ""

        if short_doc and not short_opt:
            summary = "  " + short_doc
        else:
            summary = ""

        if code_opt:
            code = "\n  defined in: {}:{}".format(stage.filename, stage.lineno)
        else:
            code = ""

        if type_opt:
            dtypes = (wrap("  inputs: ", stage.inputs) +
                      wrap("  outputs:", stage.outputs))
        else:
            dtypes = ""

        print("{name:<{width}}{summary}{description}{code}{dtypes}\n"
              "".format(name=stage.name,
                        width=name_width,
                        summary=summary,
                        code=code,
                        dtypes=dtypes,
                        description=description))



"""Exceptions raised by YMP"""

import textwrap
from inspect import stack
from typing import Optional

from click import ClickException, echo


class YmpException(Exception):
    """Base class of all YMP Exceptions"""


class YmpNoStackException(YmpException, ClickException):
    """Exception that does not lead to stack trace on CLI

    Inheriting from ClickException makes ``click`` print only the
    ``self.msg`` value of the exception, rather than allowing Python
    to print a full stack trace.

    This is useful for exceptions indicating usage or configuration
    errors. We use this, instead of `click.UsageError` and friends so
    that the exceptions can be caught and handled explicitly where
    needed.

    Note that click will call the ``show`` method on this object to
    print the exception. The default implementation from click will
    just prefix the ``msg`` with ``Error:``.

    FIXME: This does not work if the exception is raised from within
        the snakemake workflow as snakemake.snakemake catches and
        reformats exceptions.
    """


class YmpUsageError(YmpNoStackException):
    pass


class YmpSystemError(YmpNoStackException):
    """Indicates problem running YMP with available system software"""


class YmpRuleError(YmpNoStackException):
    """Indicates an error in the rules files

    This could e.g. be a Stage or Environment defined twice.

    Args:
      obj: The object causing the exception. Must have ``lineno``
        and ``filename`` as these will be shown as part of the error
        message on the command line.
      msg: The message to display
    """
    def __init__(self, obj: object, msg: str) -> None:
        self.obj = obj
        self.stack = stack()
        super().__init__(msg)

    def show(self) -> None:
        echo('Error in line %i of %s: %s' % (self.obj.lineno,
                                             self.obj.filename,
                                             self.format_message()),
             err=True)
        for fi in self.stack:
            if not fi.filename.endswith(".py"):
                echo(f"  included from {fi.filename}:{fi.lineno}")


class YmpConfigError(YmpNoStackException):
    """Indicates an error in the ymp.yml config files

    Args:
      obj: Subtree of config causing error
      msg: The message to display
      key: Key indicating part of ``obj`` causing error
      exc: Upstream exception causing error
    """
    def __init__(self, obj: object, msg: str, key: Optional[object]=None,
                 exc: Optional[Exception]=None) -> None:
        self.obj = obj
        self.stack = stack()
        self.exc = exc
        super().__init__(msg)


class YmpStageError(YmpNoStackException):
    """Indicates an error in the requested stage stack
    """
    def __init__(self, msg: str) -> None:
        super().__init__(textwrap.dedent(msg))

    def show(self) -> None:
        echo(self.format_message(), err=True)


class YmpWorkflowError(YmpNoStackException):
    """Indicates an error during workflow execution

    E.g. failures to expand dynamic variables
    """



import click
import click_completion

import ymp
from ymp.cli.env import env
from ymp.cli.make import make, submit
from ymp.cli.shared_options import group
from ymp.cli.stage import stage
from ymp.cli.show import show
from ymp.cli.init import init

click_completion.init()


def install_completion(ctx, attr, value):
    """Installs click_completion tab expansion into users shell"""
    if not value or ctx.resilient_parsing:
        return value
    shell, path = click_completion.install()
    click.echo('%s completion installed in %s' % (shell, path))
    exit(0)


def install_profiler(ctx, attr, value):
    if not value or ctx.resilient_parsing:
        return value
    import yappi
    import atexit

    def dump_profile():
        profile = yappi.get_func_stats()
        profile.sort("ttot")
        profile.print_all(out=value, columns={
            0: ("name", 120),
            1: ("ncall", 10),
            2: ("tsub", 8),
            3: ("ttot", 8),
            4: ("tavg", 8)})

    yappi.start()
    atexit.register(dump_profile)


@group()
@click.version_option(version=ymp.__version__)
@click.option(
    '--install-completion', is_flag=True,
    callback=install_completion, expose_value=False,
    help="Install command completion for the current shell. "
         "Make sure to have psutil installed.")
@click.option(
    '--profile', type=click.File(mode="w"),
    callback=install_profiler, expose_value=False,
    help="Profile execution time using Yappi"
)
def main(**kwargs):
    """
    Welcome to YMP!

    Please find the full manual at https://ymp.readthedocs.io
    """


main.add_command(make)
main.add_command(submit)
main.add_command(env)
main.add_command(stage)
main.add_command(show)
main.add_command(init)



import logging
import os
import shutil
import sys
from contextlib import ExitStack
from fnmatch import fnmatch

import click
from click import echo

import ymp
from ymp.common import ensure_list
from ymp.cli.make import snake_params, start_snakemake
from ymp.cli.shared_options import group

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

ENV_COLUMNS = ('name', 'hash', 'path', 'installed')


def get_envs(patterns=None):
    """Get environments matching glob pattern

    Args:
      envnames: list of strings to match
    """
    from ymp.env import Env
    envs = Env.get_registry()
    if patterns:
        envs = {env: envs[env] for env in envs
                if any(fnmatch(env, pat)
                       for pat in ensure_list(patterns))}
    else:
        envs = envs
    return envs


def get_env(envname):
    """Get single environment matching glob pattern

    Args:
      envname: environment glob pattern
    """
    envs = get_envs(envname)
    if not envs:
        raise click.UsageError("Environment {} unknown".format(envname))

    if len(envs) > 1:
        raise click.UsageError("Multiple environments match '{}': {}"
                               "".format(envname, envs.keys()))

    env = next(iter(envs.values()))
    if not os.path.exists(env.path):
        log.warning("Environment not yet installed")
        env.create()
    return env


@group()
def env():
    """Manipulate conda software environments

    These commands allow accessing the conda software environments managed
    by YMP. Use e.g.

    >>> $(ymp env activate multiqc)

    to enter the software environment for ``multiqc``.
    """


@env.command(name="list")
@click.option(
    "--static/--no-static", default=True,
    help="List environments statically defined via env.yml files"
)
@click.option(
    "--dynamic/--no-dynamic", default=True,
    help="List environments defined inline from rule files"
)
@click.option(
    "--all", "-a", "param_all", is_flag=True,
    help="List all environments, including outdated ones."
)
@click.option(
    "--sort", "-s", "sort_col",
    type=click.Choice(ENV_COLUMNS), default=ENV_COLUMNS[0],
    help="Sort by column"
)
@click.option(
    "--reverse", "-r", is_flag=True,
    help="Reverse sort order"
)
@click.argument("ENVNAMES", nargs=-1)
def ls(param_all, static, dynamic, sort_col, reverse, envnames):
    """List conda environments"""
    envs = get_envs(envnames)

    table_content = [
        {
            key: str(getattr(env, key))
            for key in ENV_COLUMNS
        }
        for env in envs.values()
    ]
    table_content.sort(key=lambda row: row[sort_col].upper(),
                       reverse=reverse)

    table_header = [{col: col for col in ENV_COLUMNS}]
    table = table_header + table_content
    widths = {col: max(len(row[col]) for row in table)
              for col in ENV_COLUMNS}

    lines = [" ".join("{!s:<{}}".format(row[col], widths[col])
                      for col in ENV_COLUMNS)
             for row in table]
    echo("\n".join(lines))


@env.command()
@snake_params
def prepare(**kwargs):
    "Create envs needed to build target"
    kwargs['create_envs_only'] = True
    rval = start_snakemake(kwargs)
    if not rval:
        sys.exit(1)


@env.command()
@click.option(
    "--conda-prefix", "-p",
    help="Override location for conda environments"
)
@click.option(
    "--conda-env-spec", "-e",
    help="Override conda env specs settings"
)
@click.option(
    "--dry-run", "-n", is_flag=True,
    help="Only show what would be done"
)
@click.argument("ENVNAMES", nargs=-1)
def install(conda_prefix, conda_env_spec, dry_run, envnames):
    "Install conda software environments"
    if conda_env_spec is not None:
        cfg = ymp.get_config()
        cfg.conda.env_specs = conda_env_spec

    envs = get_envs(envnames)
    log.warning(f"Creating {len(envs)} environments.")
    for env in envs.values():
        if conda_prefix:
            env.set_prefix(conda_prefix)
        env.create(dry_run)


@env.command()
@click.option(
    "--reinstall",
    help="Remove and reinstall environments rather than trying to update"
)
@click.argument("ENVNAMES", nargs=-1)
def update(envnames, reinstall):
    "Update conda environments"
    envs = get_envs(envnames)
    if reinstall:
        raise NotImplementedError("FIXME")
    log.warning(f"Updating {len(envs)} environments.")
    for env in get_envs(envnames).values():
        env.update()


@env.command()
@click.argument("ENVNAMES", nargs=-1)
def remove(envnames):
    "Remove conda environments"
    envs = get_envs(envnames)
    log.warning(f"Removing {len(envs)} environments.")
    for env in get_envs(envnames).values():
        if os.path.exists(env.path):
            log.warning("Removing %s (%s)", env.name, env.path)
            shutil.rmtree(env.path)


@env.command()
@click.option("--dest", "-d", type=click.Path(), metavar="FILE",
              help="Destination file or directory. If a directory, file names"
              " will be derived from environment names and selected export "
              "format. Default: print to standard output.")
@click.option("--overwrite", "-f", is_flag=True, default=False,
              help="Overwrite existing files")
@click.option("--create-missing", "-c", is_flag=True, default=False,
              help="Create environments not yet installed")
@click.option("--skip-missing", "-s", is_flag=True, default=False,
              help="Skip environments not yet installed")
@click.option("--filetype", "-t", type=click.Choice(['yml', 'txt']),
              help="Select export format. "
              "Default: yml unless FILE ends in '.txt'")
@click.argument("ENVNAMES", nargs=-1)
def export(envnames, dest, overwrite, create_missing, skip_missing, filetype):
    """Export conda environments

    Resolved package specifications for the selected conda
    environments can be exported either in YAML format suitable for
    use with ``conda env create -f FILE`` or in TXT format containing
    a list of URLs suitable for use with ``conda create --file
    FILE``. Please note that the TXT format is platform specific.

    If other formats are desired, use ``ymp env list`` to view the
    environments' installation path ("prefix" in conda lingo) and
    export the specification with the ``conda`` command line utlity
    directly.

    \b
    Note:
      Environments must be installed before they can be exported. This is due
      to limitations of the conda utilities.  Use the "--create" flag to
      automatically install missing environments.
    """

    envs = get_envs(envnames)

    if skip_missing and create_missing:
        raise click.UsageError(
            "--skip-missing and --create-missing are mutually exclusive")

    if dest and not filetype and dest.endswith('.txt'):
        filetype = 'txt'
    if not filetype:
        filetype = 'yml'

    missing = [env for env in envs.values() if not env.installed]
    if skip_missing:
        envs = {name: env for name, env in envs.items() if env not in missing}
    elif create_missing:
        log.warning(f"Creating {len(missing)} missing environments...")
        for n, env in enumerate(missing):
            log.warning(f"Step {n+1}/{len(missing)}:")
            env.create()
    else:
        if missing:
            raise click.UsageError(
                f"Cannot export uninstalled environment(s): "
                f"{', '.join(env.name for env in missing)}.\n"
                f"Use '-s' to skip these or '-c' to create them prior to export."
            )

    if not envs:
        if envnames and not missing:
            log.warning("Nothing to export. No environments matched pattern(s)")
        else:
            log.warning("Nothing to export")
        return

    log.warning(f"Exporting {len(envs)} environments...")

    if dest:
        if os.path.isdir(dest):
            file_names = [os.path.join(dest, ".".join((name, filetype)))
                          for name in envs.keys()]
        else:
            file_names = [dest]

        for fname in file_names:
            if not overwrite and os.path.exists(fname):
                raise click.UsageError(
                    f"File '{fname}' exists. Use '-f' to overwrite")

        with ExitStack() as stack:
            files = [stack.enter_context(open(fname, "w"))
                     for fname in file_names]
            files_stack = stack.pop_all()
    else:
        files = [sys.stdout]
        files_stack = ExitStack()
        file_names = ["stdout"]

    sep = False
    if len(files) == 1:
        sep = True
        files *= len(envs)
        file_names *= len(envs)

    with files_stack:
        generator = enumerate(zip(envs.values(), files, file_names))
        n, (env, fd, name) = next(generator)
        log.warning(f"Step {n+1}/{len(envs)}: writing to {name}")
        env.export(fd, typ=filetype)
        for n, (env, fd, name) in generator:
            log.warning(f"Step {n+1}/{len(envs)}: writing to {name}")
            if sep:
                fd.write("---\n")
            env.export(fd, typ=filetype)


@env.command()
@click.option("--all", "-a", "param_all", is_flag=True,
              help="Delete all environments")
@click.argument("ENVNAMES", nargs=-1)
def clean(param_all):
    "Remove unused conda environments"
    if param_all:  # remove up-to-date environments
        for env in ymp.env.by_name.values():
            if os.path.exists(env.path):
                log.warning("Removing %s (%s)", env.name, env.path)
                shutil.rmtree(env.path)

    # remove outdated environments
    for _, path in ymp.env.dead.items():
        log.warning("Removing (dead) %s", path)
        shutil.rmtree(path)


@env.command()
@click.argument("ENVNAME", nargs=1)
def activate(envname):
    """
    source activate environment

    Usage:
    $(ymp activate env [ENVNAME])
    """
    env = get_env(envname)
    print("source activate {}".format(env.path))


@env.command()
@click.argument("ENVNAME", nargs=1)
@click.argument("COMMAND", nargs=-1)
def run(envname, command):
    """
    Execute COMMAND with activated environment ENV

    Usage:
    ymp env run <ENV> [--] <COMMAND...>

    (Use the "--" if your command line contains option type parameters
     beginning with - or --)
    """
    env = get_env(envname)
    sys.exit(env.run(command))



"""
Module handling talking to cluster management systems

>>> python -m ymp.cluster slurm status <jobid>
"""

import subprocess as sp
import sys


def error(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class ClusterMS(object):
    pass


class Slurm(ClusterMS):
    "Talking to Slurm"

    states = {
        'BOOT_FAIL':   'failed',   # node launch failure
        'CANCELLED':   'failed',   # killed manually or by resource monitoring
        'COMPLETED':   'success',  # terminated with exit code 0
        'CONFIGURING': 'running',  # resources allocated, waiting for node
        'COMPLETING':  'running',  # some processes still active
        'DEADLINE':    'failed',   # job ran into deadline and was killed
        'FAILED':      'failed',   # job exited with code <> 0 or similar
        'NODE_FAIL':   'failed',   # job terminated due to node(s) failed
        'PENDING':     'running',  # job waiting for resources
        'PREEMPTED':   'failed',   # job terminated to make room
        'RUNNING':     'running',  # job has allocation and should be working
        'RESIZING':    'running',  # job is about to change size
        'SUSPENDED':   'running',  # job is paused
        'TIMEOUT': 'failed',       # job reached time limit
        # questionable states:
        'SPECIAL_EXIT': 'running',  # job failed but flagged "special_exit"
        'REVOKED': 'running'  # job removed due to other cluster starting it
        # Both of the above technically mean that the job isn't "done" yet,
        # running either on another cluster or on hold for a re-run after
        # initial failure.
    }

    @staticmethod
    def status(jobid):
        """Print status of job @param jobid to stdout (as needed by snakemake)

        Anectotal benchmarking shows 200ms per invocation, half used
        by Python startup and half by calling sacct. Using ``scontrol
        show job`` instead of ``sacct -pbs`` is faster by 80ms, but
        finished jobs are purged after unknown time window.
        """

        header = None
        jobs = []
        res = sp.run(['sacct', '-pbj', jobid], stdout=sp.PIPE)
        for line in res.stdout.decode('ascii').splitlines():
            line = line.strip().split("|")
            if header is None:
                header = line
                continue
            try:
                job = {key: line[header.index(key)]
                       for key in ('JobID', 'State', 'ExitCode')}
                job['snakestate'] = Slurm.states[job['State'].split(' ')[0]]
                jobs.append(job)
            except ValueError as e:
                error(e)
                error(res.stdout)
                sys.exit(1)
        snakestates = [job['snakestate'] for job in jobs]
        if 'running' in snakestates:
            print('running')
        elif 'failed' in snakestates:
            print('failed')
        else:
            print('success')
        sys.exit(0)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        error("""{self} ENGINE COMMAND ...

        Talks to cluster engine.

        Supported engines:

          - slurm

        Supported commands:

          - status <jobid> Returns running|success|failed status for
        jobid to be used by snakemake.
        """.format(self=sys.argv[0]))
        sys.exit(0 if sys.argv[1] == "-h" else 1)

    engine_name = sys.argv[1]
    command = sys.argv[2]

    if engine_name == "slurm":
        engine = Slurm()
    else:
        error("Don't know about cluster engine named '{}'".format(engine_name))
        sys.exit(1)

    if command == "status":
        if len(sys.argv) != 4:
            error("Status command needs exactly one argument, the job id.")
            sys.exit(1)
        engine.status(sys.argv[3])
    else:
        error("Command '{}' not understood.".format(command))
        sys.exit(1)



"""
Extends Snakemake Features
"""

import functools
import logging
import re
import sys
from copy import copy, deepcopy
from inspect import Parameter, signature, stack
from typing import Optional

from snakemake.exceptions import CreateRuleException, RuleException
from snakemake.io import AnnotatedString, apply_wildcards, \
    strip_wildcard_constraints
from snakemake.io import Namedlist as _Namedlist
from snakemake.rules import Rule
from snakemake.workflow import RuleInfo, Workflow

import ymp
from ymp.common import ensure_list, flatten, is_container
from ymp.exceptions import YmpRuleError
from ymp.string import ProductFormatter, make_formatter


log = logging.getLogger(__name__)  # pylint: disable=invalid-name
partial_formatter = make_formatter(partial=True, quoted=True)
partial_format = partial_formatter.format
get_names = partial_formatter.get_names


def networkx():
    import networkx
    if networkx.__version__[0] != "2":
        log.fatal("Networkx version 2.* required by YMP but {} found"
                  "".format(networkx.__version__))
        sys.exit(1)
    return networkx


def print_ruleinfo(rule: Rule, ruleinfo: RuleInfo, func=log.debug):
    """Logs contents of Rule and RuleInfo objects.

    Arguments:
      rule: Rule object to be printed
      ruleinfo: Matching RuleInfo object to be printed
      func: Function used for printing (default is log.error)
    """
    func("rule {}".format({'n': rule.name,
                           'l': rule.lineno,
                           's': rule.snakefile}))
    for attr in dir(ruleinfo):
        if attr.startswith("__"):
            continue
        func("  {}: {}".format(attr,
                               getattr(ruleinfo, attr, "")))
    func(ruleinfo.func.__code__)


def load_workflow(snakefile):
    "Load new workflow"
    return ExpandableWorkflow.load_workflow(snakefile)


def get_workflow():
    "Get active workflow, loading one if necessary"
    return ExpandableWorkflow.ensure_global_workflow()


class ExpandLateException(Exception):
    pass


class CircularReferenceException(RuleException):
    """Exception raised if parameters in rule contain a circular reference"""
    def __init__(self, deps, rule, include=None, lineno=None, snakefile=None):
        nodes = [n[0] for n in networkx().find_cycle(deps)]
        message = "Circular reference in rule {}:\n{}".format(
            rule, " => ".join(nodes + [nodes[0]]))
        super().__init__(message=message,
                         include=include,
                         lineno=lineno,
                         snakefile=snakefile,
                         rule=rule)


class InheritanceException(RuleException):
    """Exception raised for errors during rule inheritance"""
    def __init__(self, msg, rule, parent,
                 include=None, lineno=None, snakefile=None):
        message = "'{}' when deriving {} from {}".format(
            msg, rule.name, parent)
        super().__init__(message=message,
                         include=include,
                         lineno=lineno,
                         snakefile=snakefile,
                         rule=rule)


class NamedList(_Namedlist):
    """Extended version of Snakemake's :class:`~snakemake.io.Namedlist`

    - Fixes array assignment operator:
      Writing a field via ``[]`` operator updates the value accessed
      via ``.`` operator.
    - Adds ``fromtuple`` to constructor:
      Builds from Snakemake's typial ``(args, kwargs)`` tuples as
      present in ruleinfo structures.
    - Adds `update_tuple` method:
      Updates values in ``(args,kwargs)`` tuples as present in
      :class:`ruleinfo` structures.
    """
    def __init__(self, fromtuple=None, **kwargs):
        super().__init__(**kwargs)
        self._fromtuple = fromtuple
        if fromtuple:
            for value in fromtuple[0]:
                self.append(value)
            for key, value in fromtuple[1].items():
                if is_container(value):
                    start = len(self)
                    for subvalue in value:
                        self.append(subvalue)
                    self.set_name(key, start, len(self))
                else:
                    self.append(value)
                    self.add_name(key)

    def __setitem__(self, idx, value):
        # set value in list
        super().__setitem__(idx, value)
        # set value in attributes (copied references)
        for name, (i, j) in self.get_names():
            if idx >= i and (j is None or idx < j):
                self.set_name(name, i, j)

    def update_tuple(self, totuple):
        """Update values in ``(args, kwargs)`` tuple.
        The tuple must be the same as used in the constructor and
        must not have been modified.
        """
        args, kwargs = totuple
        for n, value in enumerate(args):
            if args[n] != self[n]:
                args[n] = self[n]
        for key, value in kwargs.items():
            start, end = self._names[key]
            if end:
                assert is_container(value)
                for k, j in enumerate(range(start, end)):
                    if kwargs[key][k] != self[j]:
                        kwargs[key][k] = self[j]
            else:
                if kwargs[key] != self[start]:
                    kwargs[key] = self[start]


#: describes attributes of :py:class:`snakemake.workflow.RuleInfo`
ruleinfo_fields = {
    'wildcard_constraints': {
        'format': 'argstuple',  # len(t[0]) must be == 0
    },
    'input': {
        'format': 'argstuple',
        'funcparams': ('wildcards',),
        'apply_wildcards': True,
    },
    'output': {
        'format': 'argstuple',
        'apply_wildcards': True,
    },
    'threads': {
        'format': 'int',
        'funcparams': ('wildcards', 'input', 'attempt', 'threads')
        # stored as resources._cores
    },
    'resources': {
        'format': 'argstuple',  # len(t[0]) must be == 0, t[1] must be ints
        'funcparams': ('wildcards', 'input', 'attempt', 'threads'),
    },
    'params': {
        'format': 'argstuple',
        'funcparams': ('wildcards', 'input', 'resources', 'output', 'threads'),
        'apply_wildcards': True,
    },
    'shadow_depth': {
        'format': 'string_or_true',
    },
    'priority': {
        'format': 'numeric',
    },
    'version': {
        'format': 'object',
    },
    'log': {
        'format': 'argstuple',
        'apply_wildcards': True,
    },
    'message': {
        'format': 'string',
        'format_wildcards': True,
    },
    'benchmark': {
        'format': 'string',
        'apply_wildcards': True,
    },
    'wrapper': {
        'format': 'string',
        # sets conda_env
    },
    'conda_env': {
        'format': 'string',  # path, relative to cwd or abs
        'apply_wildcards': True,
        # works only with shell/script/wrapper, not run
    },
    'singularity_img': {
        'format': 'string',
        # works ony with shell/script/wrapper, not run
    },
    'shellcmd': {
        'format': 'string',
        'format_wildcards': True
    },
    'docstring': {
        'format': 'string',
    },
    'norun': {  # does the rule have executable data?
        'format': 'bool',
    },
    'func': {
        'format': 'callable',
    },
    'script': {
        'format': 'string',
    }
    # restart_times
}


class ExpandableWorkflow(Workflow):
    """Adds hook for additional rule expansion methods to Snakemake"""
    global_workflow = None
    __expanders = []

    @classmethod
    def activate(cls):
        """Installs the ExpandableWorkflow

        Replaces the Workflow object in the snakemake.workflow module
        with an instance of this class and initializes default expanders
        (the snakemake syntax).
        """
        try:
            from snakemake.workflow import workflow
        except ImportError:
            workflow = None

        if workflow and workflow.__class__ != ExpandableWorkflow:
            # Monkey patch the existing, initialized workflow class
            workflow.__class__ = ExpandableWorkflow
            # ExpandableWorkflow.__init__ understands it may be
            # called on already initialized superclass, so this works:
            workflow.__init__()
            cls.global_workflow = workflow

    @classmethod
    def load_workflow(cls, snakefile=ymp._snakefile):
        workflow = cls(snakefile=snakefile)
        cls.global_workflow = workflow
        workflow.include(snakefile)
        return workflow

    @classmethod
    def ensure_global_workflow(cls):
        if cls.global_workflow is None:
            log.debug("Trying to activate %s...", cls.__name__)
            cls.activate()
        if cls.global_workflow is None:
            log.debug("Failed; loading default workflow")
            cls.load_workflow()
        return cls.global_workflow

    def __init__(self, *args, **kwargs):
        """Constructor for ExpandableWorkflow overlay attributes

        This may be called on an already initialized Workflow object.
        """
        # Only call super().__init__ if that hasn't happened yet
        # (as indicated by no instance attributes written to __dict__)
        if not self.__dict__:
            # only call constructor if this object hasn't been initialized yet
            super().__init__(*args, **kwargs)

        # There can only be one
        ExpandableWorkflow.global_workflow = self

        for expander in self.__expanders:
            expander.link_workflow(self)

        self._ruleinfos = {}
        self._last_rule_name = None

    @classmethod
    def register_expanders(cls, *expanders):
        """
        Register an object the expand() function of which will be called
        on each RuleInfo object before it is passed on to snakemake.
        """
        cls.__expanders = expanders
        if cls.global_workflow:
            for expander in cls.__expanders:
                expander.link_workflow(cls.global_workflow)

    @classmethod
    def clear(cls):
        if cls.global_workflow:
            cls.global_workflow = None
        # make sure there is no workflow in snakemake either
        # (we try to load that in activate())
        import snakemake.workflow
        snakemake.workflow.workflow = None

    def add_rule(self, name=None, lineno=None, snakefile=None, checkpoint=False):
        """Add a rule.

        Arguments:
          name: name of the rule
          lineno: line number within the snakefile where the rule was defined
          snakefile: name of file in which rule was defined
        """
        # super().add_rule() dynamically creates a name if `name` is None
        # stash the name so we can access it from `get_rule`
        self._last_rule_name = super().add_rule(name, lineno, snakefile)
        return self._last_rule_name

    def get_rule(self, name=None):
        """
        Get rule by name. If name is none, the last created rule is returned.

        Arguments:
          name: the name of the rule
        """
        if name is None:
            name = self._last_rule_name
        return super().get_rule(name)

    def rule(self, name=None, lineno=None, snakefile=None):
        """Intercepts "rule:"
        Here we have the entire ruleinfo object
        """
        decorator = super().rule(name, lineno, snakefile)
        rule = self.get_rule(name)

        def decorate(ruleinfo):

            if ymp.print_rule == 1:
                log.error("#### BEGIN expansion")
                print_ruleinfo(rule, ruleinfo, log.error)
                rule._ymp_print_rule = True

            for expander in reversed(self.__expanders):
                expander.expand(rule, ruleinfo)
                if ymp.print_rule == 1:
                    log.error("### expanded with " + type(expander).__name__)
                    print_ruleinfo(rule, ruleinfo, log.error)
            if ymp.print_rule:
                log.error("#### END expansion")

            # Conditionaly dump rule after YMP formatting
            if ymp.print_rule == 1:
                ymp.print_rule = 0

            # register rule with snakemake
            try:
                decorator(ruleinfo)  # does not return anything
            except AttributeError:
                print_ruleinfo(rule, ruleinfo, log.error)
                raise

        return decorate


def make_rule(name: str=None, lineno: int=None, snakefile: str=None,
              **kwargs):
    log.debug("Synthesizing rule {}".format(name))
    ruleinfo = RuleInfo(lambda: None)
    for arg in kwargs:
        setattr(ruleinfo, arg, kwargs[arg])
    ruleinfo.norun = True
    workflow = get_workflow()
    try:
        return workflow.rule(name, lineno, snakefile)(ruleinfo)
    except CreateRuleException:
        log.debug("  failed. Rule already exists?")
        return None


class BaseExpander(object):
    """
    Base class for Snakemake expansion modules.

    Subclasses should override the :meth:expand method if they need to
    work on the entire RuleInfo object or the :meth:format and
    :meth:expands_field methods if they intend to modify specific fields.
    """
    def __init__(self):
        self.workflow = None

    def link_workflow(self, workflow):
        """Called when the Expander is associated with a workflow

        May be called multiple times if a new workflow object is created.
        """
        log.debug("Linking %s with %s",
                  self.__class__.__name__, workflow.__class__.__name__)
        self.workflow = workflow

    def format(self, item, *args, **kwargs):
        """Format *item* using *\*args* and *\*\*kwargs*"""
        return item

    def format_annotated(self, item, expand_args):
        """Wrapper for :meth:format preserving *AnnotatedString* flags

        Calls :meth:format to format *item* into a new string and copies
        flags from original item.

        This is used by :meth:expand
        """
        updated = self.format(item, **expand_args)
        if isinstance(item, AnnotatedString):
            updated = AnnotatedString(updated)
            try:
                updated.flags = deepcopy(item.flags)
            except TypeError:
                log.debug(
                    "Failed to deepcopy flags for item {} with flags{}"
                    "".format(item, item.flags)
                )
            updated.flags = copy(item.flags)
        return updated

    def expands_field(self, field):
        """Checks if this expander should expand a Rule field type

        Arguments:
           field: the field to check
        Returns:
           True if *field* should be expanded.
        """
        return False

    def expand(self, rule, item, expand_args=None, rec=-1, cb=False):
        """Expands RuleInfo object and children recursively.

        Will call :meth:format (via :meth:format_annotated) on `str` items
        encountered in the tree and wrap encountered functions to be called
        once the wildcards object is available.

        Set ``ymp.print_rule = 1`` before a ``rule:`` statement in snakefiles
        to enable debug logging of recursion.

        Arguments:
          rule: The :class:snakemake.rules.Rule object to be populated with
            the data from the RuleInfo object passed from *item*
          item: The item to be expanded. Initially a
            :class:snakemake.workflow.RuleInfo object into which is recursively
            decendet. May ultimately be `None`, `str`, `function`, `int`,
            `float`, `dict`, `list` or `tuple`.
          expand_args: Parameters passed on late expansion (when the ``dag``
            tries to instantiate the `rule` into a ``job``.
          rec: Recursion level
        """
        rec = rec + 1
        if expand_args is None:
            expand_args = {}
        debug = ymp.print_rule or getattr(rule, "_ymp_print_rule", False)
        if debug:
            log.debug("{}{} {} {} in rule {} with args {}"
                      "".format(" "*rec*4, type(self).__name__,
                                type(item).__name__, item, rule, expand_args))
        if item is None:
            item = None
        elif isinstance(item, RuleInfo):
            item = self.expand_ruleinfo(rule, item, expand_args, rec)
        elif isinstance(item, str):
            item = self.expand_str(rule, item, expand_args, rec, cb)
        elif hasattr(item, '__call__'):
            item = self.expand_func(rule, item, expand_args, rec, debug)
        elif isinstance(item, int) or isinstance(item, float):
            pass
        elif isinstance(item, dict):
            item = self.expand_dict(rule, item, expand_args, rec)
        elif isinstance(item, list):
            item = self.expand_list(rule, item, expand_args, rec)
        elif isinstance(item, tuple):
            item = self.expand_tuple(rule, item, expand_args, rec)
        else:
            raise ValueError("unable to expand item '{}' with args '{}'"
                             "".format(repr(item),
                                       repr(expand_args)))

        if debug:
            log.debug("{}=> {} {}"
                      "".format(" "*(rec*4), type(item).__name__, item))

        return item

    def expand_ruleinfo(self, rule, item, expand_args, rec):
        self.current_rule = rule
        for field in filter(self.expands_field, ruleinfo_fields):
            attr = getattr(item, field)
            value = self.expand(rule, attr, expand_args=expand_args, rec=rec)
            setattr(item, field, value)
        self.current_rule = None
        return item

    def expand_str(self, rule, item, expand_args, rec, cb):
        expand_args['rule'] = rule
        try:
            return self.format_annotated(item, expand_args)
        except (KeyError, TypeError, ExpandLateException):
            # avoid recursion:
            if cb:
                raise

            def item_wrapped(wc):
                return self.expand(rule, item,
                                   expand_args={'wc': wc, 'rule': rule},
                                   cb=True)
            return item_wrapped

    def expand_func(self, rule, item, expand_args, rec, debug):
        @functools.wraps(item)
        def late_expand(*args, **kwargs):
            if debug:
                log.debug("{}{} late {} {} ".format(
                    " "*rec*4, type(self).__name__, args, kwargs))
            res = self.expand(rule, item(*args, **kwargs),
                              expand_args={'wc': args[0]}, rec=rec, cb=True)
            if debug:
                log.debug("{}=> {}".format(" "*rec*4, res))
            return res
        return late_expand

    def _make_list_wrapper(self, value):
        def wrapper(*args, **kwargs):
            res = []
            for subitem in value:
                if callable(subitem):
                    subparms = signature(subitem).parameters
                    extra_args = {
                        k: v
                        for k, v in kwargs.items()
                        if k in subparms
                    }
                    res.append(subitem(*args, **extra_args))
                else:
                    res.append(subitem)
            return res
        # Gather the arguments
        parms = tuple(set(flatten([
            list(signature(x).parameters.values())
            for x in value if callable(x)
        ])))
        # Rewrite signature
        wrapper.__signature__ = signature(wrapper).replace(parameters=parms)
        return wrapper

    def expand_dict(self, rule, item, expand_args, rec):
        for key, value in item.items():
            value = self.expand(rule, value, expand_args=expand_args, rec=rec)

            # Snakemake can't have functions in lists in dictionaries.
            # Let's fix that, even if we have to jump a lot of hoops here.
            if isinstance(value, list) and any(callable(x) for x in value):
                item[key] = self._make_list_wrapper(value)
            else:
                item[key] = value
        return item

    def expand_list(self, rule, item, expand_args, rec):
        return [self.expand(rule, subitem, expand_args=expand_args, rec=rec)
                for subitem in item]

    def expand_tuple(self, rule, item, expand_args, rec):
        return tuple(self.expand(rule, subitem, expand_args=expand_args, rec=rec)
                     for subitem in item)


class SnakemakeExpander(BaseExpander):
    """Expand wildcards in strings returned from functions.

    Snakemake does not do this by default, leaving wildcard expansion to
    the functions provided themselves. Since we never want ``{input}`` to be
    in a string returned as a file, we expand those always.
    """
    def expands_field(self, field):
        return field in ('input', 'output')

    def format(self, item, *args, **kwargs):
        if 'wc' in kwargs:
            return apply_wildcards(item, kwargs['wc'])
        return item


class FormatExpander(BaseExpander):
    """
    Expander using a custom formatter object.
    """
    regex = re.compile(
        r"""
        \{
            (?=(
                (?P<name>[^{}]+)
            ))\1
        \}
        """, re.VERBOSE)

    spec = "{{{}}}"

    def __init__(self):
        super().__init__()
        self.formatter = self.Formatter(self)

    def format(self, *args, **kwargs):
        return self.formatter.format(*args, **kwargs)

    class Formatter(ProductFormatter):
        def __init__(self, expander):
            self.expander = expander
            self.spec = expander.spec
            super().__init__()

        def parse(self, format_string):
            if format_string is None:
                return

            start = 0
            for match in self.expander.regex.finditer(format_string):
                yield (format_string[start:match.start()],
                       match.group('name'), '', None)
                start = match.end()

            yield (format_string[start:],
                   None, None, None)

    def get_names(self, pattern):
        return set(match.group('name')
                   for match in self.regex.finditer(pattern))


class ColonExpander(FormatExpander):
    """
    Expander using ``{:xyz:}`` formatted variables.
    """
    regex = re.compile(
        r"""
        \{:
            (?=(
                \s*
                 (?P<name>(?:.(?!\s*\:\}))*.)
                \s*
            ))\1
        :\}
        """, re.VERBOSE)

    spec = "{{:{}:}}"

    def __init__(self):
        super().__init__()


class RecursiveExpander(BaseExpander):
    """Recursively expands ``{xyz}`` wildcards in Snakemake rules."""
    def expands_field(self, field):
        """
        Returns true for all fields but ``shell:``, ``message:`` and
        ``wildcard_constraints``.

        We don't want to mess with the regular expressions in the fields
        in ``wildcard_constraints:``, and there is little use in expanding
        ``message:`` or ``shell:`` as these already have all wildcards applied
        just before job execution (by :meth:`format_wildcards`).
        """
        return field not in (
            'shellcmd',
            'message',
            'wildcard_constraints'
        )

    def expand(self, rule, ruleinfo):
        """Recursively expand wildcards within :class:`RuleInfo` object"""
        fields = list(filter(None.__ne__,
                             filter(self.expands_field, ruleinfo_fields)))
        # normalize field values and create namedlist dictionary
        args = {}
        for field in fields:
            attr = getattr(ruleinfo, field)
            if isinstance(attr, tuple):
                if len(attr) != 2:
                    raise Exception("Internal Error")
                # flatten named lists
                for key in attr[1]:
                    if is_container(attr[1][key]):
                        attr[1][key] = list(flatten(attr[1][key]))
                # flatten unnamed and overwrite tuples
                # also turn attr[0] into a list, making it mutable
                attr = (list(flatten(attr[0])), attr[1])

                setattr(ruleinfo, field, attr)
                args[field] = NamedList(fromtuple=attr)
            else:
                args[field] = NamedList()
                args[field].append(attr)

        # build graph of expansion dependencies
        deps = networkx().DiGraph()
        for field, nlist in args.items():
            for n, value in enumerate(nlist):
                if not isinstance(value, str):  # only strings can be expanded
                    continue
                s = "{}[{}]".format(field, n)
                # create node for value itself
                deps.add_node(s, core=True, name=field, idx=n)
                # node depends on wildcards contained in value
                deps.add_edges_from((s, t)
                                    for t in get_names(value)
                                    if t.split(".")[0].split("[")[0] in fields)
                # field node depends on all it's value nodes
                deps.add_edge(field, s)
            # create edges field.name -> field[n]
            for name, (i, j) in nlist.get_names():
                s = "{}.{}".format(field, name)
                if j is None:
                    j = i + 1
                deps.add_edges_from((s, "{}[{}]".format(field, n))
                                    for n in range(i, j))

        # sort variables so that they can be expanded in order
        try:
            nodes = list(reversed([
                node
                for node in networkx().algorithms.dag.topological_sort(deps)
                if deps.out_degree(node) > 0 and 'core' in deps.nodes[node]
            ]))
        except networkx().NetworkXUnfeasible:
            raise CircularReferenceException(deps, rule)

        # expand variables
        for node in nodes:
            var_name = deps.nodes[node]['name']
            var_idx = deps.nodes[node]['idx']
            value = args[var_name][var_idx]
            if not isinstance(value, str):
                continue

            # format what we can
            valnew = partial_format(value, **args)

            # check if any remaining wilcards refer to rule fields
            names = [re.split(r'\.|\[', name, maxsplit=1)[0]
                     for name in get_names(valnew)]
            field_names = ruleinfo_fields[var_name].get('funcparams', [])
            parm_names = [name for name in field_names if name in names]

            if parm_names:
                # Snakemake won't expand wildcards in output of functions,
                # so we need to format everything here
                def late_recursion(val, fparms):
                    def wrapper(wildcards, **kwargs):
                        # no partial here, fail if anything left
                        return strip_wildcard_constraints(val).format(
                            **kwargs, **wildcards)
                    # adjust the signature so that snakemake will pass us
                    # everything we need
                    parms = (Parameter(pname, Parameter.POSITIONAL_OR_KEYWORD)
                             for pname in fparms)
                    newsig = signature(wrapper).replace(parameters=parms)
                    wrapper.__signature__ = newsig
                    return wrapper

                valnew = late_recursion(valnew, parm_names)

            args[var_name][var_idx] = valnew

            if ymp.print_rule == 1:
                log.debug("{}::{}: {} => {}".format(rule.name,
                                                    node, value, valnew))

        # update ruleinfo
        for name in fields:
            attr = getattr(ruleinfo, name)
            if isinstance(attr, tuple):
                if len(attr) != 2:
                    raise Exception("Internal Error")
                args[name].update_tuple(attr)
            else:
                setattr(ruleinfo, name, args[name][0])


class InheritanceExpander(BaseExpander):
    """Adds class-like inheritance to Snakemake rules

    To avoid redundancy between closely related rules, e.g. rules for
    single ended and paired end data, YMP allows Snakemake rules
    to inherit from another rule.

    Example:
     .. code-block: snakemake
      rule count_reads:
        input: "{file}.R1.fq.gz", "{file}.R2.fq.gz"
        output: "{file}.readcount.txt"
        shell: "gunzip -c {input} | wc -l > {output}"

      rule count_reads_SE:  # ymp: extends count_reads
        input: "{file}.fq.gz"

    Derived rules are always created with an implicit ``ruleorder`` statement,
    making Snakemake prefer the parent rule if either parent or child rule
    could be used to generate the requested output file(s).

    Derived rules initially contain the same attributes as the parent rule.
    Each attribute assigned to the child rule overrides the matching attribute
    in the parent. Where attributes may contain named and unnamed values,
    specifying a named value overrides only the value of that name while
    specifying an unnamed value overrides all unnamed values in the parent
    attribute.
    """
    # FIXME: link to http://snakemake.readthedocs.io/en/latest/snakefiles/
    #                rules.html#handling-ambiguous-rules

    #: Comment keyword enabling inheritance
    KEYWORD = "ymp: extends"

    def __init__(self):
        super().__init__()
        self.ruleinfos = {}
        self.snakefiles = {}
        self.linemaps = None

    def get_code_line(self, rule: Rule) -> str:
        """Returns the source line defining *rule*"""
        # Load and cache Snakefile
        if rule.snakefile not in self.snakefiles:
            try:
                with open(rule.snakefile, "r") as sf:
                    self.snakefiles[rule.snakefile] = sf.readlines()
            except IOError:
                raise Exception("Can't parse ...")

        # `rule.lineno` refers to compiled code. Convert to source line number.
        if self.linemaps is None:
            self.linemaps = ExpandableWorkflow.global_workflow.linemaps
        real_lineno = self.linemaps[rule.snakefile][rule.lineno]

        return self.snakefiles[rule.snakefile][real_lineno - 1]

    def get_super(self, rule: Rule, ruleinfo: RuleInfo) -> Optional[RuleInfo]:
        """Find rule parent

        Args:
          rule: Rule object being built
          ruleinfo: RuleInfo object describing rule being built

        Returns:
          2-Tuple: name of parent rule and RuleInfo describing parent rule
          or (None, None).
        """
        self.ruleinfos[rule.name] = ruleinfo  # stash original ruleinfos

        if hasattr(ruleinfo, 'parent'):
            return ruleinfo.parent.name, self.ruleinfos[ruleinfo.parent.name]

        line = self.get_code_line(rule)

        if "#" in line:
            comment = line.split("#")[1].strip()
            if comment.startswith(self.KEYWORD):
                superrule_name = comment[len(self.KEYWORD):].strip()
                try:
                    return superrule_name, self.ruleinfos[superrule_name]
                except KeyError:
                    raise InheritanceException("Unable to find parent",
                                               rule, superrule_name)
        return None, None

    def expand(self, rule, ruleinfo):
        super_name, super_ruleinfo = self.get_super(rule, ruleinfo)
        if super_ruleinfo is None:
            return

        base_ruleinfo = deepcopy(super_ruleinfo)

        if not ruleinfo.norun:  # deriving rule is runnable, clear out base
            base_ruleinfo.shellcmd = None
            base_ruleinfo.wrapper = None
            base_ruleinfo.script = None
            base_ruleinfo.func = None
        elif not base_ruleinfo.norun:  # base is runnable, clear our deriving
            ruleinfo.norun = False
            ruleinfo.shellcmd = None
            ruleinfo.wrapper = None
            ruleinfo.script = None
            ruleinfo.func = None

        for field in dir(ruleinfo):
            if field.startswith("__") or field == "parent":
                continue

            base_attr = getattr(base_ruleinfo, field)
            override_attr = getattr(ruleinfo, field)

            if isinstance(override_attr, tuple):
                if base_attr is None:
                    base_attr = ([], {})
                if override_attr[0]:
                    base_attr = (override_attr[0], base_attr[1])
                if override_attr[1]:
                    base_attr[1].update(override_attr[1])
            elif override_attr is not None:
                base_attr = override_attr

            setattr(ruleinfo, field, base_attr)

        if not super_ruleinfo.norun:
            if super_name in self.ruleinfos:
                self.workflow.ruleorder(super_name, rule.name)


class DefaultExpander(InheritanceExpander):
    """
    Adds default values to rules

    The implementation simply makes all rules inherit from a defaults
    rule.
    """
    def __init__(self, **kwargs):
        """
        Creates DefaultExpander

        Each parameter passed is considered a RuleInfo default value. Where
        applicable, Snakemake's argtuples ``([],{})`` must be passed.
        """
        super().__init__()
        self.defaults = RuleInfo(None)
        self.defaults.norun = True

        for key, value in kwargs.items():
            setattr(self.defaults, key, value)

    def get_super(self, rule: Rule, ruleinfo: RuleInfo) -> RuleInfo:
        return ("__default__", self.defaults)


class WorkflowObject(object):
    """
    Base for extension classes defined from snakefiles

    This currently encompasses `ymp.env.Env` and `ymp.stage.Stage`.

    This mixin sets the properties ``filename`` and ``lineno`` according
    to the definition source in the rules file. It also maintains a registry
    within the Snakemake workflow object and provides an accessor method
    to this registry.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Fill filename and lineno
        # We assume the creating call is the first up the stack
        # that is not a constructor call (i.e. not __init__)
        try:
            caller = next(fi for fi in stack() if fi.function != "__init__")
            if not hasattr(self, 'filename'):
                #: str: Name of file in which object was defined
                self.filename = caller.filename
            if not hasattr(self, 'lineno'):
                #: int: Line number of object definition
                self.lineno = caller.lineno
        except IndexError:
            log.error("Failed to find source code defining %s", self)

    def register(self):
        """Add self to registry"""
        cache = self.get_registry()

        names = []
        for attr in 'name', 'altname':
            if hasattr(self, attr):
                names += ensure_list(getattr(self, attr))

        for name in names:
            if (name in cache
                and self != cache[name]
                and (self.filename != cache[name].filename
                     or self.lineno != cache[name].lineno)):
                other = cache[name]
                raise YmpRuleError(
                    self,
                    f"Failed to create {self.__class__.__name__} '{names[0]}':"
                    f" already defined in {other.filename}:{other.lineno}"
                )

        for name in names:
            cache[name] = self

    @property
    def defined_in(self):
        return self.filename

    @classmethod
    def new_registry(cls):
        return cls.get_registry(clean=True)

    @classmethod
    def get_registry(cls, clean=False):
        """
        Return all objects of this class registered with current workflow
        """
        import ymp
        cfg = ymp.get_config()
        return cfg.cache.get_cache(
            cls.__name__,
            loadfunc=ExpandableWorkflow.ensure_global_workflow,
            clean=clean)
